module Main where

import Geometry hiding (angle)
import qualified Geometry (angle)
import Types
import Control

-- import Debug.Trace
-- import IOExts(trace)

import Random(Random, randomRIO)  

-- ----------------------------------------------------------------------
--
-- State handling functions
-- 
-- ----------------------------------------------------------------------

-- One tick (time between state transitions): 30 ms
deltaT :: Double
deltaT = 30/1000

-- State of the whole system.
data State = State { car  :: Car
		   , ctrl :: Control
                   , obst :: [Figure]
		   }

-- Utility functions
setCar :: (Car-> Car)-> State-> State
setCar f s = s {car= f (car s)}

setCtrl :: (Control-> Control)-> State-> State
setCtrl f s = s {ctrl = f (ctrl s)}

setAng :: Double->State-> State
setAng a = setCtrl (setAngle a)

setAccel :: Double-> State-> State
setAccel a = setCtrl (setAcc a)

setBrk :: Bool-> State-> State
setBrk b = setCtrl (setBrake b)

initialState :: IO State
initialState = 
  do obst<- initialObstacles
     return (State { car= initialCar
		   , ctrl= initialControl
		   , obst= obst})

initialCar :: Car
initialCar =  Car {w= 0, pos= (0, 0), v= 0}


initialObstacles :: IO [Figure]
initialObstacles = 
  do nob <- randomRIO (4, 8) -- number of obstacles
     sequence (replicate nob newObst)

newObst :: IO Figure
newObst = 
  do pos <- randomPt
     cs  <- randomRIO (4, 8) -- number of corners
     pts <- mapM (\angle-> do r <- randomRIO (2.5::Double, 5)
                              return (add pos $ cart (Polar{dist=r, Geometry.angle=angle})))
                 [fromInt n*(2*pi/fromInt cs)| n<- [0..cs]]
     return $ poly pts

randomPt :: IO Point
randomPt = 
  do x<- randomRIO (-maxX, maxX)
     y<- randomRIO (-maxY, maxY)
     return (x, y)
 
-- State transition:
nextState :: State-> IO State
nextState s = return $ 
   setCar (carStateChange (acc ctr) (brake ctr) (angle ctr)) s
     where ctr= mainControl (ctrl s) (car s) (sensors (car s) (obst s))

crash :: State-> Bool
crash s = any (intersect (bodyFigure (car s))) (obst s)

-- The main loop           
loop :: Window-> State-> IO ()
loop w s =  
  do setGraphic w (drawState s)
     if crash s then return () else do
        getWindowTick w 
        evs<- getEvs
        s' <- nextState (foldl (flip procEv) s evs)
   --   putStrLn (show "Acc= "++ show (acc (ctrl s')))
        loop w $! s' where
          getEvs :: IO [Event]
          getEvs = do x<- maybeGetWindowEvent w 
                      case x of
                       Nothing -> return []
                       Just e  -> do rest <- getEvs 
                                     return (e : rest)

-- The main function, which calls the main loop with the initial state
main :: IO ()
main = runGraphics $
  do w<- openWindowEx "Haskell Racer!" Nothing 
                      (winX, winY) DoubleBuffered (Just 30)
     i<- initialState
     loop w i
     getKey w
     closeWindow w  


-- Process events:
procEv :: Event-> State-> State
procEv (Key {keysym= k, isDown=down}) 
     | k `isKey` ' ' && down     = setBrk True
     | k `isKey` ' ' && not down = setBrk False
     | isUpKey k && down         = setAccel 1
     | isUpKey k && not down     = setAccel 0
--   | isDownKey k && down       = setAcc (-1)
--   | isDownKey k && not down   = setAcc 0
procEv (Button{isLeft= True, isDown= True})  = setBrk True
procEv (Button{isLeft= True, isDown= False}) = setBrk False
procEv (MouseMove {pt= (x, y)}) = setAng (mouseToAngle x). setAccel (mouseToAcc y)
procEv _ = id

isKey :: Key-> Char-> Bool 
isKey k c = isCharKey k && keyToChar k == c

mouseToAngle :: Int-> Double
mouseToAngle x = 
  let x2= fromInt (winX `div` 2)
  in  asin ((fromInt x - x2) / x2)* phiMax / pi

mouseToAcc :: Int-> Double
mouseToAcc y = (fromInt winY - fromInt y)/fromInt winY

-- ----------------------------------------------------------------------
-- 
-- Graphics parameters and functions
-- 
-- ----------------------------------------------------------------------

-- size of the window
winX, winY :: Int
(winX, winY) = (1000, 800)

-- scale: how many pixels to one meter?
drawScale   :: Double  
drawScale   = 20

-- maximum X and Y value which still fit on the screen
maxX, maxY :: Dimension
maxX = fromInt (winX `div` 2) / drawScale 
maxY = fromInt (winY `div` 2) / drawScale

-- draw a figure with window size and scale as above
drawFig :: Figure-> Graphic
drawFig  = draw (winX `div` 2, winY `div` 2) drawScale

-- draw the state: first the car, then the speed
drawState :: State-> Graphic
drawState s = overGraphics $ 
  (if crash s then 
      [withRGB (RGB 255 0 0) $ text (winX `div` 2, winY `div` 2) "BOING!"]
   else []) ++
  [drawCar s, 
   text (10, winY-15) ("V= "++ vstr++ " km/h"),
   withRGB (RGB 140 0 0) $ 
     polygon [(10, winY- accY), (30, winY- accY), 
	      (30, winY- 10), (10, winY- 10)]] ++ 
  map drawObst (obst s) where
      accY = round(fromInt winY*(acc (ctrl s)))+ 10
      vstr = let (v1, v2) = properFraction (3.6*v(car s))
             in  show v1++ "."++ take 2 (show (truncate (v2* 100))++ "00")
      drawObst = withRGB (RGB 30 30 240). drawFig 


atCar :: Car-> Figure-> Figure
atCar c = translate (pos c). rotate (w c)

-- ----------------------------------------------------------------------
-- 
-- Behaviour and shape of the car.
-- 
-- ----------------------------------------------------------------------

-- Shape of the car: body and wheels
bodyFigure :: Car-> Figure
bodyFigure c = atCar c $ poly $ carShape

wheelsFig :: Angle-> Car-> [Figure]
wheelsFig phi c = map (atCar c. translate (-0.9, 0)) $ 
  zipWith translate
    [(0, 0.6), (1.8, 0.6), (1.8, -0.6), (0, -0.6)]
    [r, rotate phi r, rotate phi r, r] where r =rect 1 0.4

-- Draw the actual car.
drawCar :: State-> Graphic
drawCar s = 
  overGraphics $ withColor Red (drawFig (bodyFigure (car s))) : 
                 map (withRGB (RGB 50 50 50). drawFig) 
                              (wheelsFig (angle (ctrl s)) (car s))


-- Sensors: number of rays
numRays :: Double
numRays = 120

-- Calculate sensor data:
sensors :: Car-> [Figure]-> Sensor
sensors c obst =
  let sensorAngles = [(n/numRays)*2*pi | n<- [0.. numRays]]
  in  map (\ angle-> (angle, ray (pos c) angle obst)) sensorAngles


-- The central control function of the car:
-- Given acceleration, brake and steering angle, compute 
-- next state of car.
carStateChange :: Double-> Bool-> Angle-> Car-> Car
carStateChange acc brk phi (Car {w= w, pos= pos, v= v}) 
  | phi == 0  =
    let pos' = add pos (cart (Polar {Geometry.angle= w
			            , dist= v* deltaT}))
    in  Car {w= w, pos= pos', v= v'}

  | otherwise = 
    let r    = wheelBase /tan phi -- Lenkradius
        wa   = w+ phi+ pi/2  -- Lenkwinkel
        m    = add (cart (Polar{Geometry.angle= wa, dist= r})) pos    
                                        -- Lenkmittelpunkt
        zeta = v*deltaT / r -- Drehwinkel um m
        pos' = rotAround m zeta pos -- neue Pos'n
        w'   = w+ zeta -- neue Orientierung
    in  Car {w= w', pos= pos', v= v'}

    where a = frDec v + if brk then brakeDec v else acc* accMax
          v'= max 0 (v+ a*deltaT) -- Speed never goes negative.

-- refit point into window size
winMod :: Point-> Point
winMod (px, py) = (px `rMod` maxX, py `rMod` maxY) where
  rMod x m | abs x <= m = x
           | abs x >  m = if x > 0 then x- 2*m else 2*m -x

