module Geometry (
   module Graphics.HGL
 , Dimension -- Double
 , Point     -- (Dimension, Dimension)
 , Angle     -- Double
 , Figure    -- (abstract)
 , rect      -- Dimension-> Dimension-> Figure
 , triangle  -- Dimension-> Angle-> Dimension-> Figure  
 , poly      -- [Point]-> Figure
 , circle    -- Dimension-> Figure
 , translate -- Point-> Figure-> Figure
 , scale     -- Double-> Figure-> Figure
 , rotate    -- Angle-> Figure-> Figure  
 , draw      -- G.Point-> Double-> Figure-> Graphic   

 , rot       -- Angle-> Point->Point
 , rotAround -- Point-> Angle-> Point-> Point
 , add, sub  -- Point-> Point-> Point
 , intersect -- Figure-> Figure-> Bool
 , len       -- Point -> Dimension
 , distance  -- Figure-> Point-> Dimension
 , ray       -- Point-> Angle-> [Figure]-> Maybe Dimension

 , Polar(..)
 , cart
 , polar

 , fromInt
 ) where

import Graphics.HGL hiding (Point, Angle)
import qualified Graphics.HGL as G (Point, Angle) 
import List (nub, minimumBy)

-- import Debug.Trace

type Dimension = Double
type Point = (Dimension, Dimension)
type Angle = Double

rect     :: Dimension-> Dimension-> Figure
triangle :: Dimension-> Angle-> Dimension-> Figure  
poly     :: [Point]-> Figure
circle   :: Dimension-> Figure
translate :: Point-> Figure-> Figure
scale     :: Double-> Figure-> Figure
rotate    :: Angle-> Figure-> Figure  

rotAt :: Point-> Angle-> Figure-> Figure
rotAt (px, py) w = translate (px, py) .
                   rotate w . 
                   translate (-px, -py)

data Figure = Polygon [Point]
            | Circle  Point Double 

smult :: Double-> Point-> Point
smult f (x, y) 
  | f == 1    = (x, y)
  | otherwise = (f* x, f* y)

add :: Point-> Point-> Point
add (x1, y1) (x2, y2) = (x1+ x2, y1+ y2)      

sub :: Point-> Point-> Point
sub (x1, y1) (x2, y2) = (x1- x2, y1- y2) 

rot :: Angle-> Point-> Point
rot w (x, y) 
  | w == 0    = (x, y)
  | otherwise = ( x* cos w- y* sin w, 
                  x* sin w+ y* cos w) 

rotAround :: Point-> Angle-> Point-> Point
rotAround p w q = add p (rot w (sub q p))

rect a b = let x2= a/2; y2= b/2
  in Polygon [(x2, y2), (-x2, y2), (-x2, -y2), (x2, -y2), (x2, y2)] 

triangle l1 a l2 = 
  Polygon [(0, 0), (l1, 0), rot a (l2, 0), (0, 0)]

poly  ps = Polygon (checkclosed ps) where
 checkclosed [] = []
 checkclosed xs = if head xs == last xs then xs
                  else xs++ [head xs]    

circle r = Circle (0, 0) r

rotate a (Polygon ps) = Polygon (map (rot a) ps)
rotate a (Circle c r) = Circle (rot a c) r     

scale s (Polygon ps) = Polygon (map (smult s) ps)
scale s (Circle c r) = Circle  (smult s c) (r* s)

translate p (Polygon ps) = Polygon (map (add p) ps)
translate p (Circle c r) = Circle (add p c) r

draw :: G.Point-> Double-> Figure-> Graphic
draw (wx, wy) s = 
  draw'. scale s where
    draw' (Polygon ps) = 
      polygon (map (\ (x, y)-> (wx+ round x, wy+ round y)) ps) 
    draw' (Circle (x, y) r) = 
      ellipse (wx+ round (x- r), wy+ round (y- r)) 
              (wx+ round (x+ r), wy+ round (y+ r))

inC :: Point-> Double-> Point-> Bool
inC (mx, my) r (px, py) = 
    len (px- mx, py- my) <= r

len :: Point-> Double
len (x, y)= sqrt (x^2+ y^2)

det :: Point-> (Point, Point)-> Int
det (cx,cy) ((ax,ay), (bx,by)) = 
  round (signum ((by-ay)*(cx-bx)-(cy-by)*(bx-ax)))

sides :: [Point]-> [(Point, Point)]
sides []     = []
sides (x:[]) = []
sides (x:xs) = (x, head xs):(sides xs)

inP :: [Point]-> Point-> Bool
inP ps c = 
  (length. nub. map (det c). sides) ps == 1

contains :: Figure-> Point-> Bool
contains (Polygon pts)= inP pts
contains (Circle c r) = inC c r

intersect :: Figure-> Figure-> Bool
intersect (Circle (mx1, my1) r1) 
          (Circle (mx2, my2) r2)=
   len (mx2- mx1, my2- my1) <= r1+ r2

intersect (Polygon p1) (Polygon p2)= 
   any (inP p1) p2 || any (inP p2) p1

intersect (Polygon p) (Circle c r)= 
   inP p c || any (inC c r) p
intersect (Circle c r) (Polygon p)=
   inP p c || any (inC c r) p


data Polar = Polar { angle :: Angle, dist  :: Dimension }        deriving (Show, Eq)    
cart :: Polar-> Point
cart p = rot (angle p) (dist p, 0)

polar :: Point-> Polar
polar (x, y)
  | (x, y) == (0, 0) = Polar {dist= r, angle = 0} -- beliebig
  | otherwise        = Polar {dist= r, angle= asin (y/r)} 
                       where r = len (x, y)


distance :: Point-> Figure-> Dimension
distance p (Circle c r) =
    abs (len (sub c p) - r)
distance q (Polygon ps) 
  | inP ps q  = 0 -- q is in the polygon
  | otherwise = minimum (map (abs. distPointToLine q) (sides ps))

-- So how does this work...
-- Der Winkel ang ist die Steigung der durch (p1, p2) beschriebenen Geraden g.
-- Wir rotieren gegen den Winkel ang, damit ist g parallel zur Ordinate
-- (x-Achse). Jetzt machen wir eine Fallunterscheidung:
--  1. p1 liegt links von p2, dann:
--  i) q liegt links von p1
--     dann ist die Entfernung q zu (p1, p2) die Entfernung von q zu p1;
--  ii) q liegt rechts von p2
--      dann ist die Entfernung q zu (p1, p2) die Entfernung von q zu p2;
--  iii) q liegt zwischen p1 und p2
--       dann ist die Entfernung von q zu (p1, p2) die Distanzen der 
--       y-Koordinaten von q und p1 (oder p2-- ist dasselbe). 
--  2. p2 liegt rechts von p1, dann i)... ii)... iii)... analog
--  Die links/rechts Entscheidungen sind die Vergleiche der X-Koordinaten.
--  Uff.
distPointToLine :: Point-> (Point, Point)-> Dimension
distPointToLine q (p1, p2) =
  let ang = angle (polar (sub p2 p1))
      [p1', p2', q']  = map (rot (- ang)) [p1, p2, q]
  in  if fst p1' <= fst p2' then
        (if fst q' < fst p1' then len (sub p1 q)
         else (if fst p2' < fst q' then len (sub p2 q)
                                   else snd q'-snd p1'))
      else -- p2' < p1'
        (if fst q' < fst p2' then len (sub p2 q)
         else (if fst p1' < fst q' then len (sub p1 q)
                                   else snd q'- snd p1'))


-- Shoot a ray from point p with angle a, check wether it hits one of the 
-- figures and if so return minimal distance.
ray :: Point-> Angle-> [Figure]-> Maybe Dimension
ray p a figs = 
  minimumBy compMaybe $ 
    map (\f-> rayX (rotate (-a) (translate (smult (-1) p) f))) figs

-- Compare Maybes, putting Nothings greater then everything else
compMaybe :: Ord a=> Maybe a-> Maybe a-> Ordering
compMaybe (Just a) (Just b) = compare a b
compMaybe Nothing _  = GT
compMaybe _ Nothing  = LT

-- A total minimum function.
minMaybe :: Ord a=> [a]-> Maybe a
minMaybe [] = Nothing
minMaybe xs = Just (minimum xs)


-- Shoot ray along X axis (positive only), check wether it hits figure.
rayX :: Figure-> Maybe Dimension
rayX (Polygon ps) =
  -- For polygons, find all sides which cross the X axis, and then the one
  -- with the minimum distance (but larger/equal than 0)
  minMaybe $ filter (0<=) $ map distOnX (filter crossesX (sides ps))
--  let crossingSides = map distOnX (filter crossesX (sides ps))
--  in if null crossingSides then Nothing else Just (minimum crossingSides)
rayX (Circle (cx, cy) r) =
  -- Too tired to be exact now, this is a good lower bound for wee circles :-)
  if cy <= r && cx >= r then Just (cx- r) else Nothing

-- Does the Line (p1, p2) cross the X axis? It does if the y sign changes. 
crossesX :: (Point, Point)-> Bool
crossesX (p1, p2) = signum (snd p1) /= signum (snd p2)

-- Given two points, at which point does the line described by them cross
-- the X axis? 
-- These two points p1, p2 describe a line with g = \lambda(p2- p1)+ p1,
-- hence gx = \lambda(p2x- p1x)+ p1x (similar for gy).
-- Solve this for gy= 0, then calculate gx.
distOnX :: (Point, Point)-> Dimension
distOnX (p1, p2) =
  let lambda = - (snd p1)/(snd p2- snd p1)
  in  lambda*(fst p2- fst p1)+ fst p1


-- A useful wee function
fromInt :: Num a=> Int-> a
fromInt n = fromInteger (toInteger n)


{- -- Test data.

let p = poly  [(-1,1), (-1,2), (1,2), (1,1)]
let q = (3.0,3.0)

map (\a-> ray q a [p]) [fromInt n/10*2*pi | n<- [0..10]]

-}
